# -*- coding: future_fstrings -*-

from enum import Enum


class GroupUserAccessRight(Enum):
    NONE = 'None'
    MEMBER = 'Member'
    ADMIN = 'Admin'
    CONTRIBUTOR = 'Contributor'
    VIEWER = 'Viewer'


class PrincipalType(Enum):
    USER = "User"
    GROUP = "Group"
    APP = "App"


class DatasourceUserAccessRight(Enum):
    # Removes permission to access the datasource
    NONE = 'None'
    # Datasets owned by the user have read access to this datasource
    READ = 'Read'
    # The user can override the effective identity for PowerBI Embedded
    READ_OVERRIDE_EFFECTIVE_IDENTITY = 'ReadOverrideEffectiveIdentity'


class CredentialType(Enum):
    ANONYMOUS = 'Anonymous'
    BASIC = 'Basic'
    KEY = 'Key'
    OAUTH2 = 'OAuth2'
    WINDOWS = 'Windows'


class EncryptedConnection(Enum):
    ENCRYPTED = 'Encrypted'
    NOT_ENCRYPTED = 'NotEncrypted'


class EncryptionAlgorithm(Enum):
    NONE = 'None'
    RSA_OAEP = 'RSA-OAEP'


class PrivacyLevel(Enum):
    NONE = 'None'
    PUBLIC = 'Public'
    ORGANIZATIONAL = 'Organizational'
    PRIVATE = 'Private'



# -*- coding: future_fstrings -*-
import json
from typing import Dict, Union, Optional

from .base import Deserializable
from .enums import CredentialType, DatasourceUserAccessRight, PrincipalType, EncryptedConnection, EncryptionAlgorithm, \
    PrivacyLevel


class GatewayPublicKey(Deserializable):
    exponent_key = 'exponent'
    modulus_key = 'modulus'

    def __init__(
            self,
            exponent: str,
            modulus: str
    ):
        """Constructs a GatewayPublicKey object

        :param exponent: The exponent of the public key
        :param modulus: The modulus of the public key
        """
        self.exponent = exponent
        self.modulus = modulus

    def as_dict(self) -> Dict[str, str]:
        return {
            self.exponent_key: self.exponent,
            self.modulus_key: self.modulus
        }

    @classmethod
    def from_dict(cls, dictionary: Dict[str, str]) -> 'GatewayPublicKey':
        """Constructs a GatewayPublicKey from a dictionary

        :param dictionary: the dictionary describing the GatewayPublicKey
        :return: GatewayPublicKey based on the dictionary
        :rtype: GatewayPublicKey
        """
        exponent = dictionary.get(cls.exponent_key)
        modulus = dictionary.get(cls.modulus_key)

        return cls(exponent, modulus)

    def __repr__(self) -> str:
        return f'<GatewayPublicKey exponent={self.exponent} modulus={self.modulus}>'


class Gateway(Deserializable):
    id_key = 'id'
    name_key = 'name'
    type_key = 'type'
    gateway_annotation_key = 'gatewayAnnotation'
    public_key_key = 'publicKey'
    status_key = 'gatewayStatus'

    def __init__(
            self,
            gateway_id: str,
            name: str,
            gateway_type: str,
            gateway_annotation: str,
            public_key: GatewayPublicKey,
            status: str
    ):
        """Constructs a Gateway object

        :param gateway_id: The gateway id
        :param name: The gateway name
        :param gateway_type: The gateway type
        :param gateway_annotation: Gateway metadata in json format
        :param public_key: The gateway public key
        :param status: The gateway connectivity status
        """
        self.id = gateway_id
        self.name = name
        self.type = gateway_type
        self.gateway_annotation = gateway_annotation
        self.public_key = public_key
        self.status = status

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Union[str, Dict[str, str]]]) -> 'Gateway':
        """Constructs a Gateway object from a dict

        :param dictionary: Dictionary describing the gateway
        :return: Gateway based on the dictionary
        :rtype: Gateway
        """
        gateway_id = dictionary.get(cls.id_key)
        if gateway_id is None:
            raise RuntimeError("Gateway dictionary has no id key")

        name = dictionary.get(cls.name_key)
        gateway_type = dictionary.get(cls.type_key)
        gateway_annotation = dictionary.get(cls.gateway_annotation_key)
        public_key = GatewayPublicKey.from_dict(dictionary.get(cls.public_key_key))
        status = dictionary.get(cls.status_key)

        return cls(gateway_id, name, gateway_type, gateway_annotation, public_key, status)

    def __repr__(self) -> str:
        return f'<Gateway id={self.id} name={self.name}>'


class GatewayDatasource(Deserializable):
    gateway_datasource_id_key = 'id'
    gateway_id_key = 'gatewayId'
    credential_type_key = 'credentialType'
    datasource_name_key = 'datasourceName'
    datasource_type_key = 'datasourceType'
    connection_details_key = 'connectionDetails'

    def __init__(
            self,
            gateway_datasource_id: str,
            gateway_id: str,
            credential_type: CredentialType,
            datasource_name: str,
            datasource_type: str,
            connection_details: str
    ):
        """Constructs a GatewayDatasource object

        :param gateway_datasource_id: The unique id for this datasource
        :param gateway_id: The associated gateway id
        :param credential_type: Type of datasource credentials
        :param datasource_name: The name of the datasource
        :param datasource_type: The type of the datasource
        :param connection_details: Connection details in json format
        """
        self.id = gateway_datasource_id
        self.gateway_id = gateway_id
        self.credential_type = credential_type
        self.datasource_name = datasource_name
        self.datasource_type = datasource_type
        self.connection_details = connection_details

    @classmethod
    def from_dict(cls, dictionary: Dict[str, str]) -> 'GatewayDatasource':
        """Constructs a GatewayDatasource object from a dict

        :param dictionary: Dictionary describing the gatewayDatasource
        :return: GatewayDatasource based on the dictionary
        :rtype: GatewayDatasource
        """
        gateway_datasource_id = dictionary.get(cls.gateway_datasource_id_key)
        if gateway_datasource_id is None:
            raise RuntimeError("GatewayDatasource dictionary has no id key")

        gateway_id = dictionary.get(cls.gateway_id_key)
        # use round brackets below to access enum by value
        credential_type = CredentialType(dictionary.get(cls.credential_type_key))
        datasource_name = dictionary.get(cls.datasource_name_key)
        datasource_type = dictionary.get(cls.datasource_type_key)
        connection_details = json.dumps(dictionary.get(cls.connection_details_key))

        return cls(
            gateway_datasource_id,
            gateway_id,
            credential_type,
            datasource_name,
            datasource_type,
            connection_details
        )

    def __repr__(self):
        return f'<GatewayDatasource id={self.id} name={self.datasource_name} type={self.datasource_type}>'


class DatasourceUser(Deserializable):
    datasource_access_right_key = 'datasourceAccessRight'
    email_address_key = 'emailAddress'
    display_name_key = 'displayName'
    identifier_key = 'identifier'
    principal_type_key = 'principalType'

    def __init__(
        self,
        datasource_access_right: DatasourceUserAccessRight,
        email_address: str = "",
        display_name: str = "",
        identifier: str = "",
        principal_type: Optional[PrincipalType] = None
    ):
        """Constructs a DataSourceUser object

        :param datasource_access_right: The user access rights for the datasource
        :param email_address: Email address of the user
        :param display_name: Display name of the principal
        :param identifier: Identifier of the principal
        :param principal_type: The principal type
        """
        self.datasource_access_right = datasource_access_right
        self.email_address = email_address
        self.display_name = display_name
        self.identifier = identifier
        self.principal_type = principal_type

    @classmethod
    def from_dict(cls, dictionary: Dict[str, str]) -> 'DatasourceUser':
        datasource_user_id = dictionary.get(cls.identifier_key)
        if datasource_user_id is None:
            raise RuntimeError("DatasourceUser dictionary has no identifier key")

        datasource_user_access_right = DatasourceUserAccessRight(dictionary.get(cls.datasource_access_right_key))
        email_address = dictionary.get(cls.email_address_key, "")
        display_name = dictionary.get(cls.display_name_key, "")
        principal_type_value = dictionary.get(cls.principal_type_key, None)
        principal_type = PrincipalType(principal_type_value) if principal_type_value is not None \
            else principal_type_value

        return cls(datasource_user_access_right, email_address, display_name, datasource_user_id, principal_type)

    def as_set_values_dict(self) -> Dict[str, str]:
        set_values_dict = dict()

        set_values_dict[self.datasource_access_right_key] = self.datasource_access_right.value

        if self.email_address:
            set_values_dict[self.email_address_key] = self.email_address

        if self.display_name:
            set_values_dict[self.display_name_key] = self.display_name

        if self.identifier:
            set_values_dict[self.identifier_key] = self.identifier

        if self.principal_type:
            set_values_dict[self.principal_type_key] = self.principal_type.value

        return set_values_dict

    def __repr__(self) -> str:
        return f'<DatasourceUser id={self.identifier} type={self.principal_type.name} display_name={self.display_name}>'


class DatasourceConnectionDetails:
    server_key = "server"
    database_key = "database"
    url_key = "url"

    def __init__(
        self,
        server: str = "",
        database: str = "",
        url: str = ""
    ):
        """Creates a DatasourceConnectionDetails object

        :param server: The connection server
        :param database: The connection database
        :param url: The connection url
        """
        self.server = server
        self.database = database
        self.url = url

    def __repr__(self):
        server_part = f'server={self.server}' if self.server else self.server
        database_part = f'database={self.database}' if self.database else self.database
        url_part = f'url={self.url}' if self.url else self.url

        return f'<DatasourceConnectionDetails {server_part} {database_part} {url_part}>'

    def as_set_values_dict(self) -> Dict[str, str]:
        """Returns a dictionary with only those values of attributes that are set

        :return: Dictionary with set values
        """
        set_values = dict()

        if self.server:
            set_values[self.server_key] = self.server

        if self.database:
            set_values[self.database_key] = self.database

        if self.url:
            set_values[self.url_key] = self.url

        return set_values

    def to_json(self) -> str:
        """Provides a json string that can be used in a PublishDatasourceToGatewayRequest

        :return: json string of set values
        """
        json_dict = self.as_set_values_dict()

        # dump to json string
        # remove spaces between object keys and values
        # replace double backslashes with single slashes
        # remove double quotes at start and end of str
        return json.dumps(json_dict) \
            .replace(r'": ', r'":') \
            .replace('\\\\', '\\')


class CredentialDetails:
    credentials_key = "credentials"
    credential_type_key = "credentialType"
    encrypted_connection_key = "encryptedConnection"
    encryption_algorithm_key = "encryptionAlgorithm"
    privacy_level_key = "privacyLevel"
    use_caller_aad_identity_key = "useCallerAADIdentity"
    use_end_user_o_auth_2_credentials_key = "useEndUserOAuth2Credentials"

    def __init__(
        self,
        credentials: str,
        credential_type: CredentialType,
        encrypted_connection: EncryptedConnection,
        encryption_algorithm: EncryptionAlgorithm,
        privacy_level: PrivacyLevel,
        use_caller_aad_identity: Optional[bool] = None,
        use_end_user_o_auth_2_credentials: Optional[bool] = None
    ):
        """Constructs a CredentialsDetails object

        :param credentials: The credentials to access a datasource
        :param credential_type: The type of credentials to access a datasource
        :param encrypted_connection: Encryption behaviour applied to the datasource connection
        :param encryption_algorithm: The encryption algorithm. For cloud datasource, use 'None'. For an on-premises
         datasource, use the gateway public key with the 'RSA-OAEP' algorithm.
        :param privacy_level: The privacy level. This becomes relevant when combining data from multiple datasources.
        :param use_caller_aad_identity: Should the caller's AAD identity be used for OAuth2 credentials configuration
        :param use_end_user_o_auth_2_credentials: Should the end-user's OAuth2 credentials be used for connecting to
         the datasource in DirectQuery mode. Only supported for Direct Query to SQL Azure.
        """
        self.credentials = credentials
        self.credential_type = credential_type
        self.encrypted_connection = encrypted_connection
        self.encryption_algorithm = encryption_algorithm
        self.privacy_level = privacy_level
        self.use_caller_aad_identity = use_caller_aad_identity
        self.use_end_user_o_auth_2_credentials = use_end_user_o_auth_2_credentials

    def to_dict(self):
        return {
            self.credentials_key: self.credentials,
            self.credential_type_key: self.credential_type.value,
            self.encrypted_connection_key: self.encrypted_connection.value,
            self.encryption_algorithm_key: self.encryption_algorithm.value,
            self.privacy_level_key: self.privacy_level.value,
            self.use_caller_aad_identity_key: self.use_caller_aad_identity,
            self.use_end_user_o_auth_2_credentials_key: self.use_end_user_o_auth_2_credentials
        }

    def __repr__(self) -> str:
        return f"<CredentialDetails type={self.credential_type.value}>"


class PublishDatasourceToGatewayRequest:
    datasource_type_key = "dataSourceType"
    connection_details_key = "connectionDetails"
    credential_details_key = "credentialDetails"
    datasource_name_key = "datasourceName"

    def __init__(
        self,
        datasource_type: str,
        connection_details: str,
        credential_details: CredentialDetails,
        datasource_name: str
    ):
        """Constructs a PublishDatasourceToGatewayRequest

        :param datasource_type: The datasource type
        :param connection_details: The connection details
        :param credential_details: The credentials to access the datasource
        :param datasource_name: The datasource name
        """
        self.datasource_type = datasource_type
        self.connection_details = connection_details
        self.credential_details = credential_details
        self.datasource_name = datasource_name

    def to_dict(self) -> Dict[str, Union[str, Dict[str, str]]]:
        return {
            self.datasource_type_key: self.datasource_type,
            self.connection_details_key: self.connection_details,
            self.credential_details_key: self.credential_details.to_dict(),
            self.datasource_name_key: self.datasource_name
        }

    def __repr__(self):
        return '<PublishDatasourceToGatewayRequest ' \
               f'name={self.datasource_name} ' \
               f'type={self.datasource_type}>'



# -*- coding: future_fstrings -*-
from typing import List, Type, Any

import requests
import json

from requests.exceptions import HTTPError

from .gateway import Gateway, GatewayDatasource, DatasourceUser, PublishDatasourceToGatewayRequest
from .base import Deserializable


class Gateways:
    # url snippets
    gateways_snippet = 'gateways'
    datasources_snippet = 'datasources'
    users_snippet = 'users'

    # json keys
    odata_response_wrapper_key = 'value'

    def __init__(self, client):
        self.client = client
        self.base_url = f'{self.client.api_url}/{self.client.api_version_snippet}/{self.client.api_myorg_snippet}'

    def get_gateways(self) -> List[Gateway]:
        """Fetches all gateways the user is an admin for"""

        # form the url
        url = f'{self.base_url}/{self.gateways_snippet}'

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Gateways request returned http error: {response.json()}')

        return self._models_from_get_multiple_response(response, Gateway)

    def get_gateway(self, gateway_id: str) -> Gateway:
        """Return the specified gateway

        :param gateway_id: The gateway id
        :return: The gateway
        """
        # form the url
        url = f'{self.base_url}/{self.gateways_snippet}/{gateway_id}'

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Gateway request returned http error: {response.json()}')

        return self._model_from_get_one_response(response, Gateway)

    def get_datasources(self, gateway_id: str) -> List[GatewayDatasource]:
        """Returns a list of datasources from the specified gateway

        :param gateway_id: The gateway id to return responses for
        :return: list
            The list of datasources
        """

        # form the url
        url = f'{self.base_url}/{self.gateways_snippet}/{gateway_id}/{self.datasources_snippet}'

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Gateway Datasources request returned http error: {response.json()}')

        return self._models_from_get_multiple_response(response, GatewayDatasource)

    def get_datasource_users(self, gateway_id: str, datasource_id: str) -> List[DatasourceUser]:
        """Returns a list of users who have access to the specified datasource

        :param gateway_id: The gateway id
        :param datasource_id: The datasource id
        """
        # form the url
        url = f'{self.base_url}/{self.gateways_snippet}/{gateway_id}' \
              f'/{self.datasources_snippet}/{datasource_id}/{self.users_snippet}'

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Datasource Users request returned http error: {response.json()}')

        return self._models_from_get_multiple_response(response, DatasourceUser)

    def create_datasource(
        self,
        gateway_id: str,
        datasource_to_gateway_request: PublishDatasourceToGatewayRequest
    ) -> GatewayDatasource:
        """Creates a new datasource on the specified gateway

        :param gateway_id: The gateway id
        :param datasource_to_gateway_request: Request describing the datasource to be created
        """
        # form the url
        url = f'{self.base_url}/{self.gateways_snippet}/{gateway_id}/{self.datasources_snippet}'

        # define request body
        body = datasource_to_gateway_request.to_dict()

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.post(url, headers=headers, json=body)

        # 201 is the only successful code, raise an exception on any other response code
        if response.status_code != 201:
            raise HTTPError(f'Create Datasource request returned the following http error: {response.json()}')

        return self._model_from_get_one_response(response, GatewayDatasource)

    def delete_datasource(
        self,
        gateway_id: str,
        datasource_id: str
    ) -> None:
        """Deletes the specified datasource from the specified gateway

        :param gateway_id: The gateway id
        :param datasource_id: The datasource id
        :return: None
        """
        # form the url
        url = f'{self.base_url}/{self.gateways_snippet}/{gateway_id}/{self.datasources_snippet}/{datasource_id}'

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.delete(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(f'Delete Datasource request returned the following http error: {response.json()}')

        return None

    def add_datasource_user(
        self,
        gateway_id: str,
        datasource_id: str,
        datasource_user: DatasourceUser
    ) -> None:
        """Grants or updates the permissions required to use the specified datasource for the specified user. Note:
        This method does not work with a service principal, only with a username password flow, for which the user
        has given consent.

        :param gateway_id: The gateway id
        :param datasource_id: The datasource id
        :param datasource_user: The datasource user to add
        """
        # form the url
        url = f'{self.base_url}/{self.gateways_snippet}/{gateway_id}/{self.datasources_snippet}/{datasource_id}/' \
              f'{self.users_snippet}'

        # define request body
        body = datasource_user.as_set_values_dict()

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.post(url, headers=headers, json=body)

        if response.status_code != 200:
            # add datasource user requests return an empty body; get the error from headers instead
            raise HTTPError(f'Add datasource user request returned the following http error: {response.json()} '
                            f'with status code: {response.status_code}')

        return None

    @classmethod
    def _models_from_get_multiple_response(
        cls,
        response: requests.Response,
        model_class: Type[Deserializable]
    ) -> List[Any]:
        """Creates a list of models from a http response object

        :param response:
            The http response object
        :param model_class:
            The model to transform the response items into
        :return: list
            The list of model_class instances
        """

        # parse json response into a dict
        response_dict = json.loads(response.text)

        # Add parsed Gateway objects to list
        items = []
        for entry in response_dict[cls.odata_response_wrapper_key]:
            items.append(model_class.from_dict(entry))

        return items

    @classmethod
    def _model_from_get_one_response(
        cls,
        response: requests.Response,
        model_class: Type[Deserializable]
    ) -> Any:
        # parse
        response_dict = json.loads(response.text)

        return model_class.from_dict(response_dict)



# -*- coding: future_fstrings -*-
import requests
import json

from requests.exceptions import HTTPError


class Features:
    # url snippets
    features_snippet = 'availableFeatures'

    def __init__(self, client):
        self.client = client
        self.base_url = f'{self.client.api_url}/{self.client.api_version_snippet}/{self.client.api_myorg_snippet}'

    @property
    def embed_trial(self):
        feat = self.get_available_features(feature_name='embedTrial')
        feat.usage = feat.additional_info['usage']
        return feat

    @property
    def automatically_push_app_to_end_user(self):
        return self.get_available_features(feature_name='automaticallyPushAppToEndUsers')

    @property
    def publish_app_to_entire_organization(self):
        return self.get_available_features(feature_name='publishAppToEntireOrganization')

    def get_available_features(self, feature_name=None):
        # If feature_name is none, returns all the available features
        if feature_name is not None:
            feature_part = f"(featureName='{feature_name}')"
        else:
            feature_part = ''

        # form the url
        url = f"{self.base_url}/{self.features_snippet}{feature_part}"
        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Datasets request returned http error: {response.json()}')

        return Features.features_from_get_available_features_response(response)

    @staticmethod
    def features_from_get_available_features_response(response):
        response_dict = json.loads(response.text)

        # check wether we are returning a single feature or a list of features
        if 'features' in response_dict:
            features_list = []
            for feature in response_dict['features']:
                additional_info = feature[Feature.additional_info_key] if Feature.additional_info_key in feature else  None
                features_list.append(Feature(
                    feature[Feature.name_key],
                    feature[Feature.state_key],
                    feature[Feature.extended_state_key],
                    additional_info
                    ))
            return features_list
        else:
            additional_info = response_dict[Feature.additional_info_key] if Feature.additional_info_key in response_dict else  None
            return Feature(
                response_dict[Feature.name_key],
                response_dict[Feature.state_key],
                response_dict[Feature.extended_state_key],
                additional_info
                )


class Feature:
    name_key = 'name'
    state_key = 'state'
    extended_state_key = 'extendedState'
    additional_info_key = 'additionalInfo'

    def __init__(self, name, state, extended_state, additional_info=None):
        self.name = name
        self.state = state
        self.extended_state = extended_state

        if additional_info is not None:
            self.additional_info = additional_info



# -*- coding: future_fstrings -*-
import requests

import datetime

from requests.exceptions import HTTPError


class ActivityLogs:

    def __init__(self, client):
        self.client = client
        self.base_url = f'{self.client.api_url}/{self.client.api_version_snippet}/{self.client.api_myorg_snippet}'

        self.activities_events_snippet = "activityevents"
        self.group_part = "admin"   # This is always admin. Not really a group, but follows the
                                    # format of the rest of the library code

    def get_activity_logs(self, st, et=None, filter=None):
        """
        Get's the activity log for the specified date or date range. If et is None, it will get all logs (from midnight
        to 11:59:59 UTC) for the date specified by st. If et is set, it will retrieve logs from st-et. Note that the
        Power BI Activity Service currently supports only retrieving one day of logs at a time.

        "filter" is a string parameter that's sent to the service to filter the types of events returned. For example,
        "Activity eq 'viewreport' and UserId eq 'john@contoso.com'" gets report views for john(contoso.com).
        Right now the service only supports the operators ['eq', 'and'].

        NOTE: It appears that only data from December 15th, 2019 and on can be retrieved by the API as of the writing
        of this code. This isn't an official limitation I've found in the documentation, but seems to be the case.

        NOTE: This API allows at most 200 Requests per hour.

        For a good overview of the service, see https://powerbi.microsoft.com/en-us/blog/the-power-bi-activity-log-makes-it-easy-to-download-activity-data-for-custom-usage-reporting/

        :param st: The date to retrieve usage for (python datetime).
        :param et: The date to retrieve usage for (python datetime).
        :param filter: A string that defines a filter for retrieving the information. See the Power BI REST API
                       Documentation for details.
        :return:
        """
        # TODO: It would be nice if the available parameters for the "filter" function were defined somewhere in code.

        if et is None:
            dt_str = st.strftime("%Y-%m-%d")
            st_dt_str = f"{dt_str}T00:00:00"
            et_dt_str = f"{dt_str}T23:59:59"
        else:
            st_dt_str = st.strftime("%Y-%m-%dT:%H%M%S")
            et_dt_str = et.strftime("%Y-%m-%dT:%H%M%S")

        # https://api.powerbi.com/v1.0/myorg/admin/activityevents?startDateTime='{st_dt_str}'&endDateTime='{et_dt_str}'

        # form the url
        filter_snippet = f"startDateTime='{st_dt_str}'&endDateTime='{et_dt_str}'"
        url = f'{self.base_url}/{self.group_part}/{self.activities_events_snippet}?{filter_snippet}'

        if filter is not None:
            url += f"$filter={filter}"

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Datasets request returned http error: {response.json()}')

        response_obj = response.json()

        event_entities = response_obj["activityEventEntities"]
        continuation_uri = response_obj["continuationUri"]
        continuation_token = response_obj["continuationToken"]

        activity_events = event_entities

        # Even if nothing is returned, it takes around 24 tries until no continuation token is returned.
        # (This is how Microsoft says the API is to be used.)
        # It seems to send the first set of actual data around 12-15 calls in. This doesn't seem to change even if you
        # slow down the API calls (in total number of calls required or when the first set of actual data is returned).

        cont_count = 1
        while continuation_token is not None:

            response = requests.get(continuation_uri, headers=headers)
            response_obj = response.json()

            event_entities = response_obj["activityEventEntities"]
            continuation_uri = response_obj["continuationUri"]
            continuation_token = response_obj["continuationToken"]

            activity_events.extend(event_entities)
            # print(f"{cont_count}: {len(event_entities)}")
            cont_count += 1

        # print(f"Took {cont_count} tries to exhaust continuation token for {len(activity_events)} events.")

        # Convert Datetime Strings to Python datetimes
        _date_fmt_str = '%Y-%m-%dT%H:%M:%S'
        for event in activity_events:
            event["CreationTime"] = datetime.datetime.strptime(event["CreationTime"], _date_fmt_str)
            # Change the Timezone to UTC
            event["CreationTime"] = event["CreationTime"].replace(tzinfo=datetime.timezone.utc)

        return activity_events



# -*- coding: future_fstrings -*-
import requests
import json
from pypowerbi.utils import convert_datetime_fields

from requests.exceptions import HTTPError
from .dataset import *


class Datasets:
    # url snippets
    groups_snippet = 'groups'
    datasets_snippet = 'datasets'
    tables_snippet = 'tables'
    rows_snippet = 'rows'
    parameters_snippet = 'parameters'
    set_parameters_snippet = 'Default.UpdateParameters'
    bind_gateway_snippet = 'Default.BindToGateway'
    refreshes_snippet = 'refreshes'
    refresh_schedule_snippet = 'refreshSchedule'

    # json keys
    get_datasets_value_key = 'value'

    def __init__(self, client):
        self.client = client
        self.base_url = f'{self.client.api_url}/{self.client.api_version_snippet}/{self.client.api_myorg_snippet}'

    def count(self, group_id=None):
        """
        Evaluates the number of datasets
        :param group_id: The optional group id
        :return: The number of datasets as returned by the API
        """
        return len(self.get_datasets(group_id))

    def has_dataset(self, dataset_id, group_id=None):
        """
        Evaluates if the dataset exists
        :param dataset_id: The id of the dataset to evaluate
        :param group_id: The optional group id
        :return: True if the dataset exists, False otherwise
        """
        datasets = self.get_datasets(group_id)

        for dataset in datasets:
            if dataset.id == str(dataset_id):
                return True

        return False

    def get_datasets(self, group_id=None):
        """
        Fetches all datasets
        https://msdn.microsoft.com/en-us/library/mt203567.aspx
        :param group_id: The optional group id to get datasets from
        :return: The list of the datasets found
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}'
        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Datasets request returned http error: {response.json()}')

        return self.datasets_from_get_datasets_response(response)

    def get_dataset(self, dataset_id, group_id=None):
        """
        Gets a single dataset
        https://msdn.microsoft.com/en-us/library/mt784653.aspx
        :param dataset_id: The id of the dataset to get
        :param group_id: The optional id of the group to get the dataset from
        :return: The dataset returned by the API
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}'
        # form the headers
        headers = self.client.auth_header
        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Datasets request returned http error: {response.json()}')

        return Dataset.from_dict(json.loads(response.text))

    def post_dataset(self, dataset, group_id=None):
        """
        Posts a single dataset
        https://msdn.microsoft.com/en-us/library/mt203562.aspx
        :param dataset: The dataset to push
        :param group_id: The optional group id to push the dataset to
        :return: The pushed dataset as returned by the API
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}'
        # form the headers
        headers = self.client.auth_header
        # form the json dict
        json_dict = DatasetEncoder().default(dataset)

        # get the response
        response = requests.post(url, headers=headers, json=json_dict)

        # 201 - Created. The request was fulfilled and a new Dataset was created.
        if response.status_code != 201:
            raise HTTPError(response, f'Post Datasets request returned http code: {response.json()}')

        return Dataset.from_dict(json.loads(response.text))

    def delete_dataset(self, dataset_id, group_id=None):
        """
        Deletes a dataset
        :param dataset_id: The id of the dataset to delete
        :param group_id: The optional group id to delete the dataset from
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}'
        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.delete(url, headers=headers)

        # 200 is the only successful code
        if response.status_code != 200:
            raise HTTPError(response, f'Delete Dataset request returned http error: {response.json()}')

    def delete_all_datasets(self, group_id=None):
        """
        Deletes all datasets
        :param group_id: The optional group id of the group to delete all datasets from
        """
        # get all the datasets and delete each one
        datasets = self.get_datasets(group_id)
        for dataset in datasets:
            self.delete_dataset(group_id, dataset.id)

    def get_tables(self, dataset_id, group_id=None):
        """
        Gets tables from a dataset
        https://msdn.microsoft.com/en-us/library/mt203556.aspx
        :param dataset_id: The id of the dataset which to get tables from
        :param group_id: The optional id of the group which to get tables from
        :return: A list of tables from the given group and dataset
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/{self.tables_snippet}'
        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Datasets request returned http error: {response.json()}')

        return self.tables_from_get_tables_response(response)

    def put_table(self, dataset_id, table_name, table, group_id=None):
        """
        Updates the metadata and schema for the specified table within the specified dataset from "My Workspace".
        https://docs.microsoft.com/en-us/rest/api/power-bi/pushdatasets/datasets_puttable
        :param dataset_id: The id of the dataset to put the table in
        :param table_name: The name of the table to put
        :param table: The table object to update
        :param group_id: The optional id of the group to put the table in
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

            # form the url
            url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/' \
                  f'{self.tables_snippet}/{table_name}'
            # form the headers
            headers = self.client.auth_header
            # form the json dict
            json_dict = TableEncoder().default(table)

            # get the response
            response = requests.post(url, headers=headers, json=json_dict)

            # 200 is the only successful code
            if response.status_code != 200:
                raise HTTPError(response, f'Post row request returned http error: {response.json()}')

    def post_rows(self, dataset_id, table_name, rows, group_id=None):
        """
        Posts rows to a table in a given dataset
        https://msdn.microsoft.com/en-us/library/mt203561.aspx
        :param dataset_id: The id of the dataset to post rows to
        :param table_name: The name of the table to post rows to
        :param rows: The rows to post to the table
        :param group_id: The optional id of the group to post rows to
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/' \
              f'{self.tables_snippet}/{table_name}/{self.rows_snippet}'
        # form the headers
        headers = self.client.auth_header
        # form the json dict
        row_encoder = RowEncoder()
        json_dict = {
            'rows': [row_encoder.default(x) for x in rows]
        }

        # get the response
        response = requests.post(url, headers=headers, json=json_dict)

        # 200 is the only successful code
        if response.status_code != 200:
            raise HTTPError(response, f'Post row request returned http error: {response.json()}')

    def delete_rows(self, dataset_id, table_name, group_id=None):
        """
        Deletes all rows from a table in a given dataset
        https://msdn.microsoft.com/en-us/library/mt238041.aspx
        :param dataset_id: The id of the dataset to delete the rows from
        :param table_name: The name of the table to delete the rows from
        :param group_id: The optional id of the group to delete the rows from
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/' \
              f'{self.tables_snippet}/{table_name}/{self.rows_snippet}'
        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.delete(url, headers=headers)

        # 200 is the only successful code
        if response.status_code != 200:
            raise HTTPError(response, f'Post row request returned http error: {response.json()}')

    def get_dataset_parameters(self, dataset_id, group_id=None):
        """
        Gets all parameters for a single dataset
        https://msdn.microsoft.com/en-us/library/mt784653.aspx
        :param dataset_id: The id of the dataset from which you want the parameters
        :param group_id: The optional id of the group to get the dataset's parameters
        :return: The dataset parameters returned by the API
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/{self.parameters_snippet}'
        # form the headers
        headers = self.client.auth_header
        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Dataset parameters request returned http error: {response.json()}')

        return json.loads(response.text)

    def set_dataset_parameters(self, dataset_id, params, group_id=None):
        """
        Sets parameters for a single dataset
        https://docs.microsoft.com/en-gb/rest/api/power-bi/datasets/updateparametersingroup
        :param dataset_id: The id of the dataset which you want to update
        :param params: Dict of parameters to set on the dataset
        :param group_id: The optional id of the group to get the dataset's parameters
        :return: The dataset parameters returned by the API
        """
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/{self.set_parameters_snippet}'

        update_details = [{"name": k, "newValue": str(v)} for k, v in params.items()]
        body = {"updateDetails": update_details}

        headers = self.client.auth_header

        response = requests.post(url, headers=headers, json=body)

        if response.status_code != 200:
            raise HTTPError(response, f'Setting dataset parameters failed with http error: {response.json()}')

    def refresh_dataset(self, dataset_id, notify_option=None, group_id=None):
        """
        Refreshes a single dataset
        :param dataset_id: The id of the dataset to refresh
        :param notify_option: The optional notify_option to add in the request body
        :param group_id: The optional id of the group
        """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/{self.refreshes_snippet}'

        # form the headers
        headers = self.client.auth_header

        if notify_option is not None:
            json_dict = {
                'notifyOption': notify_option
            }
        else:
            json_dict = None

        # get the response
        response = requests.post(url, headers=headers, json=json_dict)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 202:
            raise HTTPError(response, f'Refresh dataset request returned http error: {response.json()}')

    def get_dataset_gateway_datasources(self, dataset_id, group_id=None):
        """
                Gets the gateway datasources for a dataset
                :param dataset_id: The id of the dataset
                :param group_id: The optional id of the group
                """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}{self.datasets_snippet}/{dataset_id}/datasources'

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            print(url)
            raise HTTPError(response, f'Dataset gateway datasources request returned http error: {response.json()}')

        data_sources = json.loads(response.text)["value"]

        return data_sources

    def bind_dataset_gateway(self, dataset_id, gateway_id, group_id=None):
        """
                Binds a dataset to a gateway
                https://docs.microsoft.com/en-gb/rest/api/power-bi/datasets/bindtogatewayingroup
                :param dataset_id: The id of the dataset
                :param gateway_id: The id of the gateway
                :param group_id: The optional id of the group
                """
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/{self.bind_gateway_snippet}'

        body = {"gatewayObjectId": gateway_id}
        headers = self.client.auth_header

        response = requests.post(url, headers=headers, json=body)

        if response.status_code != 200:
            raise HTTPError(response, f'Binding gateway to dataset failed with http error: {response.json()}')

    def get_dataset_refresh_history(self, dataset_id, group_id=None, top=None):
        """
                Gets the refresh history of a dataset
                :param dataset_id: The id of the dataset to refresh
                :param group_id: The optional id of the group
                :param top: The number of refreshes to retrieve. 5 will get the last 5 refreshes.
                """
        # group_id can be none, account for it
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/{self.refreshes_snippet}'

        if top is not None:
            url = f'{url}?$top={top}'

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Dataset refresh history request returned http error: {response.json()}')

        refresh_data = json.loads(response.text)["value"]

        # Convert the date strings into datetime objects
        time_fields = ['startTime', 'endTime']
        refresh_data = convert_datetime_fields(refresh_data, time_fields)

        return refresh_data

    def update_refresh_schedule(
        self,
        dataset_id: str,
        refresh_schedule: RefreshSchedule,
        group_id: Optional[str] = None
    ):
        """Updates the refresh schedule for a given dataset in a given workspace.

        :param dataset_id: The dataset id
        :param refresh_schedule: The updates for the refresh schedule. If a field remains None, no changes are made.
        :param group_id: The workspace id of the workspace in which the dataset resides. If None, 'My Workspace' is
        assumed.
        """
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/{self.refresh_schedule_snippet}'

        # form the headers
        headers = self.client.auth_header

        # form the body
        body = RefreshScheduleRequest(refresh_schedule).as_dict()

        # get the response
        response = requests.patch(url, headers=headers, json=body)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(f'Update refresh schedule request returned the following http error:{response.json()}')

    def get_refresh_schedule(self, dataset_id: str, group_id: Optional[str] = None):
        """Retrieves the refresh schedule for a given dataset and group id

        :param dataset_id: The dataset id for which the refresh schedule should be retrieved
        :param group_id:  The group in which the dataset resides. If None 'My Workspace' is used.
        """
        if group_id is None:
            groups_part = '/'
        else:
            groups_part = f'/{self.groups_snippet}/{group_id}/'

        # form the url
        url = f'{self.base_url}{groups_part}/{self.datasets_snippet}/{dataset_id}/{self.refresh_schedule_snippet}'

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(f'Get refresh schedule request returned the following http error:{response.json()}')

        return self.refresh_schedule_from_get_refresh_schedule_response(response)


    @classmethod
    def datasets_from_get_datasets_response(cls, response):
        """
        Creates a list of datasets from a http response object
        :param response: The http response object
        :return: A list of datasets created from the given http response object
        """
        # load the response into a dict
        response_dict = json.loads(response.text)
        datasets = []
        # go through entries returned from API
        for entry in response_dict[cls.get_datasets_value_key]:
            datasets.append(Dataset.from_dict(entry))

        return datasets

    @classmethod
    def tables_from_get_tables_response(cls, response):
        """
        Creates a list of tables from a http response object
        :param response: The http response object
        :return: A list of tables created from the given http response object
        """
        # load the response into a dict
        response_dict = json.loads(response.text)
        tables = []
        # go through entries returned from API
        for entry in response_dict[cls.get_datasets_value_key]:
            tables.append(Table.from_dict(entry))

        return tables

    @classmethod
    def refresh_schedule_from_get_refresh_schedule_response(cls, response: requests.Response):
        response_dict = json.loads(response.text)
        return RefreshSchedule.from_dict(response_dict)



# -*- coding: future_fstrings -*-
import json
from enum import Enum
from typing import List, Optional, Dict, Union


class Dataset:
    # json keys
    id_key = 'id'
    name_key = 'name'
    add_rows_api_enabled_key = 'addRowsAPIEnabled'
    configured_by_key = 'configuredBy'
    is_refreshable_key = 'isRefreshable'
    is_effective_identity_required_key = 'isEffectiveIdentityRequired'
    is_effective_identity_roles_required_key = 'isEffectiveIdentityRolesRequired'
    is_on_prem_gateway_required_key = 'isOnPremGatewayRequired'
    tables_key = 'tables'

    def __init__(self, name, dataset_id=None, tables=None, add_rows_api_enabled=None,
                 configured_by=None, is_refreshable=None, is_effective_identity_required=None,
                 is_effective_identity_roles_required=None, is_on_prem_gateway_required=None):
        self.name = name
        self.id = dataset_id
        self.tables = tables
        self.add_rows_api_enabled = add_rows_api_enabled
        self.configured_by = configured_by
        self.is_refreshable = is_refreshable
        self.is_effective_identity_required = is_effective_identity_required
        self.is_effective_identity_roles_required = is_effective_identity_roles_required
        self.is_on_prem_gateway_required = is_on_prem_gateway_required

    @classmethod
    def from_dict(cls, dictionary):
        """
        Creates a dataset from a dictionary, key values for 'id' and 'name' required
        :param dictionary: The dictionary to create the dataset from
        :return: A dataset created from the given dictionary
        """
        # id is required
        if Dataset.id_key in dictionary:
            dataset_id = str(dictionary[Dataset.id_key])
            # id cannot be whitespace
            if dataset_id.isspace():
                raise RuntimeError('Dataset dict has empty id key value')
        else:
            raise RuntimeError('Dataset dict has no id key')
        # name is required
        if Dataset.name_key in dictionary:
            dataset_name = str(dictionary[Dataset.name_key])
            # name cannot be whitespace
            if dataset_id.isspace():
                raise RuntimeError('Dataset dict has empty name key value')
        else:
            raise RuntimeError('Dataset dict has no name key')

        # add api enabled is optional
        if Dataset.add_rows_api_enabled_key in dictionary:
            add_rows_api_enabled = bool(dictionary[Dataset.add_rows_api_enabled_key])
        else:
            add_rows_api_enabled = None

        # configured by is optional
        if Dataset.configured_by_key in dictionary:
            configured_by = str(dictionary[Dataset.configured_by_key])
        else:
            configured_by = None

        # is refreshable is optional
        if Dataset.is_refreshable_key in dictionary:
            is_refreshable = bool(dictionary[Dataset.is_refreshable_key])
        else:
            is_refreshable = None

        # is effective identity required is optional
        if Dataset.is_effective_identity_required_key in dictionary:
            is_effective_identity_required = bool(dictionary[Dataset.is_effective_identity_required_key])
        else:
            is_effective_identity_required = None

        # is effective identity roles required is optional
        if Dataset.is_effective_identity_roles_required_key in dictionary:
            is_effective_identity_roles_required = bool(dictionary[Dataset.is_effective_identity_roles_required_key])
        else:
            is_effective_identity_roles_required = None

        # is on prem gateway required is optional
        if Dataset.is_on_prem_gateway_required_key in dictionary:
            is_on_prem_gateway_required = bool(dictionary[Dataset.is_on_prem_gateway_required_key])
        else:
            is_on_prem_gateway_required = None

        return Dataset(dataset_name, dataset_id, add_rows_api_enabled=add_rows_api_enabled,
                       configured_by=configured_by, is_refreshable=is_refreshable,
                       is_effective_identity_required=is_effective_identity_required,
                       is_effective_identity_roles_required=is_effective_identity_roles_required,
                       is_on_prem_gateway_required=is_on_prem_gateway_required)

    def __repr__(self):
        return f'<Dataset {str(self.__dict__)}>'


class DatasetEncoder(json.JSONEncoder):
    def default(self, o):
        table_encoder = TableEncoder()

        json_dict = {
            Dataset.name_key: o.name,
            Dataset.tables_key: [table_encoder.default(x) for x in o.tables],
        }

        return json_dict


class Table:
    name_key = 'name'
    columns_key = 'columns'
    measures_key = 'measures'
    rows_key = 'rows'

    @classmethod
    def from_dict(cls, dictionary):
        """
        Creates a table from a dictionary, 'name' key value required
        :param dictionary: The dictionary to create the table from
        :return: A table created from the dictionary
        """
        # name is required
        if Table.name_key in dictionary:
            table_name = str(dictionary[Table.name_key])
            # name cannot be whitespace
            if table_name.isspace():
                raise RuntimeError('Table dict has empty name key value')
        else:
            raise RuntimeError('Table dict has no name key')

        # columns are optional
        if Table.columns_key in dictionary:
            table_columns = [Table.from_dict(x) for x in dictionary[Table.columns_key]]
        else:
            table_columns = None

        # measures are optional
        if Table.measures_key in dictionary:
            table_measures = [Table.from_dict(x) for x in dictionary[Table.measures_key]]
        else:
            table_measures = None

        return Table(name=table_name, columns=table_columns, measures=table_measures)

    def __init__(self, name, columns, measures=None, rows=None):
        self.name = name
        self.columns = columns
        self.measures = measures
        self.rows = rows

    def __repr__(self):
        return f'<Table {str(self.__dict__)}>'


class TableEncoder(json.JSONEncoder):
    def default(self, o):
        json_dict = {
            Table.name_key: o.name,
        }

        if o.columns is not None:
            column_encoder = ColumnEncoder()
            json_dict[Table.columns_key] = [column_encoder.default(x) for x in o.columns]

        if o.measures is not None:
            measure_encoder = MeasureEncoder()
            json_dict[Table.measures_key] = [measure_encoder.default(x) for x in o.measures]

        if o.rows is not None:
            row_encoder = RowEncoder()
            json_dict[Table.row_key] = [row_encoder.default(x) for x in o.rows]

        return json_dict


class Measure:
    name_key = 'name'
    expression_key = 'expression'
    formatstring_key = 'formatString'
    is_hidden_key = 'isHidden'

    @classmethod
    def from_dict(cls, dictionary):
        # name is required
        if Measure.name_key in dictionary:
            measure_name = str(dictionary[Measure.name_key])
            # name cannot be whitespace
            if measure_name.isspace():
                raise RuntimeError('Measure dict has empty name key value')
        else:
            raise RuntimeError('Measure dict has no name key')

        # expression is required
        if Measure.expression_key in dictionary:
            measure_expression = str(dictionary[Measure.expression_key])
            # expression cannot be whitespace
            if measure_expression.isspace():
                raise RuntimeError('Measure dict has empty expression key value')
        else:
            raise RuntimeError('Measure dict has no expression key')

        if Measure.formatstring_key in dictionary:
            measure_formatstring = str(dictionary[Measure.formatstring_key])
        else:
            measure_formatstring = None

        if Measure.is_hidden_key in dictionary:
            measure_is_hidden = bool(dictionary[Measure.is_hidden_key])
        else:
            measure_is_hidden = None

        return Measure(name=measure_name, expression=measure_expression, formatstring=measure_formatstring,
                       is_hidden=measure_is_hidden)

    def __init__(self, name, expression, formatstring=None, is_hidden=None):
        self.name = name
        self.expression = expression
        self.formatstring = formatstring
        self.is_hidden = is_hidden

    def __repr__(self):
        return f'<Measure {str(self.__dict__)}>'


class MeasureEncoder(json.JSONEncoder):
    def default(self, o):
        json_dict = {
            Measure.name_key: o.name,
            Measure.expression_key: o.expression,
        }

        if o.formatstring is not None:
            json_dict[Measure.formatstring_key] = o.formatstring

        if o.is_hidden is not None:
            json_dict[Measure.is_hidden_key] = o.is_hidden

        return json_dict


class Column:
    name_key = 'name'
    datatype_key = 'dataType'
    formatstring_key = 'formatString'

    def __init__(self, name, data_type, formatstring=None):
        self.name = name
        self.data_type = data_type
        self.formatstring = formatstring

    def __repr__(self):
        return f'<Column {str(self.__dict__)}>'


class ColumnEncoder(json.JSONEncoder):
    def default(self, o):
        json_dict = {
            Column.name_key: o.name,
            Column.datatype_key: o.data_type
        }

        if o.formatstring is not None:
            json_dict[Column.formatstring_key] = o.formatstring

        return json_dict


class Row:
    def __init__(self, **kwargs):
        for key in kwargs:
            setattr(self, key, kwargs[key])

    def __repr__(self):
        return f'<Row {str(self.__dict__)}>'


class RowEncoder(json.JSONEncoder):
    def default(self, o):
        return o.__dict__


class ScheduleNotifyOption(Enum):
    MAIL_ON_FAILURE = "MailOnFailure"
    NO_NOTIFICATION = "NoNotification"


class RefreshSchedule:
    notify_option_key = 'NotifyOption'
    days_key = 'days'
    enabled_key = 'enabled'
    local_time_zone_id_key = 'localTimeZoneId'
    times_key = 'times'

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Union[str, bool]]) -> 'RefreshSchedule':
        notify_option_value = dictionary.get(cls.notify_option_key, None)
        notify_option = ScheduleNotifyOption(notify_option_value) if notify_option_value else None
        days = dictionary.get(cls.days_key, None)
        enabled = dictionary.get(cls.enabled_key, None)
        local_time_zone_id = dictionary.get(cls.local_time_zone_id_key, "")
        times = dictionary.get(cls.times_key, None)

        return cls(notify_option, days, enabled, local_time_zone_id, times)

    def __init__(
        self,
        notify_option: Optional[ScheduleNotifyOption] = None,
        days: Optional[List[str]] = None,
        enabled: Optional[bool] = None,
        local_time_zone_id: Optional[str] = "",
        times: Optional[List[str]] = None
    ):
        """Constructs a RefreshSchedule object

        :param notify_option: Notify on failure or not
        :param days: Days to execute the refresh
        :param enabled: Is the scheduled refresh enabled
        :param local_time_zone_id: The id of the timezone to use. Follows the conventions described here:
        https://docs.microsoft.com/en-us/dotnet/api/system.timezoneinfo.id
        :param times: Times to schedule the refresh on each specified day
        """

        self.notify_option = notify_option
        self.days = days
        self.enabled = enabled
        self.local_time_zone_id = local_time_zone_id
        self.times = times

    def as_set_values_dict(self) -> Dict[str, Union[str, bool, List[str]]]:
        set_values_dict = dict()

        if self.notify_option:
            set_values_dict[self.notify_option_key] = self.notify_option.value

        if self.days:
            set_values_dict[self.days_key] = self.days

        if self.enabled is not None:
            set_values_dict[self.enabled_key] = self.enabled

        if self.local_time_zone_id:
            set_values_dict[self.local_time_zone_id_key] = self.local_time_zone_id

        if self.times:
            set_values_dict[self.times_key] = self.times

        return set_values_dict

    def __repr__(self):
        return f'<RefreshSchedule {str(self.as_set_values_dict())}>'


class RefreshScheduleRequest:
    value_key = "value"

    def __init__(self, refresh_schedule: RefreshSchedule):
        """Constructs an update Refresh Schedule request

        :param refresh_schedule: The desired refresh schedule
        """
        self.refresh_schedule = refresh_schedule

    def as_dict(self) -> Dict[str, Dict[str, Union[str, bool, List[str]]]]:
        return {
            self.value_key: self.refresh_schedule.as_set_values_dict()
        }

    def __repr__(self):
        return f'<RefreshScheduleRequest {str(self.as_dict())}>'



# -*- coding: future_fstrings -*-
import requests
import json
import urllib.parse

from requests.exceptions import HTTPError
from .group import Group
from .group_user import GroupUser


class Groups:
    # url snippets
    groups_snippet = 'groups'
    users_snippet = 'users'

    # json keys
    get_reports_value_key = 'value'

    def __init__(self, client):
        self.client = client
        self.base_url = f'{self.client.api_url}/{self.client.api_version_snippet}/{self.client.api_myorg_snippet}'

    def create_group(self, name, workspace_v2=False):
        """Creates a new workspace

        :param name: The name of the new group to create
        :param workspace_v2: Create a workspace V2
        :return: Group
            The newly created group
        """
        # validate request body
        if name is None or name == "":
            raise ValueError("Group name cannot be empty or None")

        # define request body
        body = {'name': name}

        # create url
        url = f'{self.base_url}/{self.groups_snippet}'

        uri_parameters = []

        if workspace_v2:
            stripped_workspace_v2 = json.dumps(workspace_v2).strip('"')
            uri_parameters.append(f'workspaceV2={urllib.parse.quote(stripped_workspace_v2)}')

        # add query parameters to url if any
        if len(uri_parameters) > 0:
            url += f'?{str.join("&", uri_parameters)}'

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.post(url, headers=headers, json=body)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(f'Add group request returned the following http error: {response.json()}')

        return self.create_group_from_create_group_response(response)

    @staticmethod
    def create_group_from_create_group_response(response):
        """Creates a Group object from the response to a create_group call

        :param response:
            The http response object
        :return:
            Group object describing the newly created group
        """
        group_dict = json.loads(response.text)
        return Group.from_dict(group_dict)

    def add_group_user(self, group_id, group_user):
        """Adds a user to a group

        :param group_id:
            str - id of the group to add the user to
        :param group_user:
            GroupUser - Description of the user that should be added to the group
        :return:
            None
        """
        # validate request body
        if not isinstance(group_user, GroupUser):
            raise TypeError("group_user should be of type group_user.GroupUser !")

        # define request body
        body = group_user.as_set_values_dict()

        # create url
        stripped_group_id = json.dumps(group_id).strip('"')
        url = f'{self.base_url}/{self.groups_snippet}/{urllib.parse.quote(stripped_group_id)}/{self.users_snippet}'

        # form the headers
        headers = self.client.auth_header

        # get the response
        response = requests.post(url, headers=headers, json=body)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            # add group user requests return an empty body; get the error from headers instead
            error_info = response.headers['x-powerbi-error-info']
            raise HTTPError(f'Add group request returned the following http error: {error_info}')

    def count(self):
        """
        Evaluates the number of groups that the client has access to
        :return: int
            The number of groups
        """
        return len(self.get_groups())

    def has_group(self, group_id):
        """
        Evaluates if client has access to the group
        :param group_id:
        :return: bool
            True if the client has access to the group, False otherwise
        """
        groups = self.get_groups()

        for group in groups:
            if group.id == str(group_id):
                return True

        return False

    def get_groups(self, filter_str=None, top=None, skip=None):
        """
        Fetches all groups that the client has access to
        :param filter_str: OData filter string to filter results
        :param top: int > 0, OData top parameter to limit to the top n results
        :param skip: int > 0,  OData skip parameter to skip the first n results
        :return: list
            The list of groups
        """
        query_parameters = []

        if filter_str:
            query_parameters.append(f'$filter={urllib.parse.quote(filter_str)}')

        if top:
            stripped_top = json.dumps(top).strip('"')
            query_parameters.append(f'$top={urllib.parse.quote(stripped_top)}')

        if skip:
            stripped_skip = json.dumps(skip).strip('"')
            query_parameters.append(f'$skip={urllib.parse.quote(stripped_skip)}')

        # form the url
        url = f'{self.base_url}/{self.groups_snippet}'

        # add query parameters to url if any
        if len(query_parameters) > 0:
            url += f'?{str.join("&", query_parameters)}'

        # form the headers
        headers = self.client.auth_header
        # get the response
        response = requests.get(url, headers=headers)

        # 200 is the only successful code, raise an exception on any other response code
        if response.status_code != 200:
            raise HTTPError(response, f'Get Groups request returned http error: {response.json()}')

        return self.groups_from_get_groups_response(response)

    @classmethod
    def groups_from_get_groups_response(cls, response):
        """
        Creates a list of groups from a http response object
        :param response:
            The http response object
        :return: list
            The list of groups
        """
        # load the response into a dict
        response_dict = json.loads(response.text)
        groups = []

        # go through entries returned from API
        for entry in response_dict[cls.get_reports_value_key]:
            groups.append(Group.from_dict(entry))

        return groups



from .client import *
from .dataset import *
from .datasets import *
from .report import *
from .reports import *
from .imports import *
from .groups import *
from .gateways import *
from .gateway import *



from typing import Optional, Dict, List, Union
import json

from pypowerbi import CredentialType


class CredentialsBase:
    CREDENTIAL_TYPE: Optional[CredentialType] = None
    credential_data_key = "credentialData"

    def __init__(self):
        self.credential_data: Dict[str, Union[str, List[Dict[str, str]]]] = {
            self.credential_data_key: []
        }

    def add_credential_data(self, key, value):
        self.credential_data[self.credential_data_key].append(
            {"name": key, "value": value}
        )

    def to_json(self) -> str:
        return json.dumps(self.credential_data, separators=(',', ':'))\
            .replace('\\\\', '\\')


class UsernamePasswordCredentials(CredentialsBase):
    username_key = "username"
    password_key = "password"

    def __init__(self, username: str, password: str):
        super().__init__()

        if not username:
            raise ValueError("An empty string is not a valid username!")
        if not password:
            raise ValueError("An empty string is not a valid password!")

        self.username = username
        self.password = password

        super().add_credential_data(self.username_key, username)
        super().add_credential_data(self.password_key, password)


class AnonymousCredentials(CredentialsBase):
    CREDENTIAL_TYPE = CredentialType.ANONYMOUS

    def __init__(self):
        super().__init__()
        self.credential_data[self.credential_data_key] = ""

    def __repr__(self) -> str:
        return f'<AnonymousCredentials>'


class BasicCredentials(UsernamePasswordCredentials):
    CREDENTIAL_TYPE = CredentialType.BASIC

    def __init__(self, username: str, password: str):
        super().__init__(username, password)

    def __repr__(self) -> str:
        return f'<BasicCredentials username={self.username}>'


class KeyCredentials(CredentialsBase):
    CREDENTIAL_TYPE = CredentialType.KEY
    key_key = "key"

    def __init__(self, key: str):
        super().__init__()

        if not key:
            raise ValueError("An empty string is not a valid key!")

        self.key = key

        super().add_credential_data(self.key_key, key)

    def __repr__(self) -> str:
        return f'<KeyCredentials>'


class OAuth2Credentials(CredentialsBase):
    CREDENTIAL_TYPE = CredentialType.OAUTH2
    access_token_key = "accessToken"

    def __init__(self, access_token: str):
        super().__init__()

        if not access_token:
            raise ValueError("An empty string is not a valid access token!")

        super().add_credential_data(self.access_token_key, access_token)

    def __repr__(self) -> str:
        return f'<OAuth2Credentials>'


class WindowsCredentials(UsernamePasswordCredentials):
    CREDENTIAL_TYPE = CredentialType.WINDOWS

    def __init__(self, username: str, password: str):
        super().__init__(username, password)

    def __repr__(self):
        return f'<WindowsCredentials username={self.username}>'



# -*- coding: future_fstrings -*-
class GroupUser:
    group_user_access_right_key = 'groupUserAccessRight'
    email_address_key = 'emailAddress'
    display_name_key = 'displayName'
    identifier_key = 'identifier'
    principal_type_key = 'principalType'

    def __init__(
        self,
        group_user_access_right,
        email_address="",
        display_name="",
        identifier="",
        principal_type=None
    ):
        """Constructs a GroupUser object

        :param group_user_access_right: Enum GroupUserAccessRight - The access right to assign to the GroupUser
        :param email_address: str - E-mail address of the user if principal type is user
        :param display_name: str - Display name of the principal
        :param identifier: str - Identifier of the principal
        :param principal_type: Enum PrincipalType - The principal type
        """
        self.group_user_access_right = group_user_access_right
        self.email_address = email_address
        self.display_name = display_name
        self.identifier = identifier
        self.principal_type = principal_type

    def as_set_values_dict(self):
        """Convert GroupUser object to dict with only values that are actually set. This dict can be used for
        groups.add_group_user requests.

        :return: Dict with object attributes in camelCase as keys, and attribute values as values.
        """
        group_user_dict = dict()

        if self.group_user_access_right:
            group_user_dict[self.group_user_access_right_key] = self.group_user_access_right.value

        if self.email_address:
            group_user_dict[self.email_address_key] = self.email_address

        if self.display_name:
            group_user_dict[self.display_name_key] = self.display_name

        if self.identifier:
            group_user_dict[self.identifier_key] = self.identifier

        if self.principal_type:
            group_user_dict[self.principal_type_key] = self.principal_type.value

        return group_user_dict



import abc
from typing import Dict, Union


class Deserializable(metaclass=abc.ABCMeta):
    """Interface to ensure operations modules need fewer methods to turn responses into objects"""
    @classmethod
    @abc.abstractmethod
    def from_dict(cls, dictionary: Dict[str, Union[str, Dict[str, str]]]):
        pass



# -*- coding: future_fstrings -*-
import datetime


"""
This file contains helper and utility functions used elsewhere in the library.
"""


# Datetime's come in from PowerBI in the format 2019-03-05T03:09:31.493Z
_date_fmt_str = '%Y-%m-%dT%H:%M:%S.%fZ'
_date_fmt_str2 = '%Y-%m-%dT%H:%M:%SZ'


def date_from_powerbi_str(dstr):
    """
    Utility function to convert datetime strings from the Power BI service into Python Datetime objects

    :param dstr: A String retrieved from the Power BI Service that's a datetime
    :return: A Python datetime object generated from the parameter
    """
    if "." in dstr:
        return datetime.datetime.strptime(dstr, _date_fmt_str)
    else:
        # Fractional seconds are not zero padded in the API and will not be included at all if 0, thus the second format
        return datetime.datetime.strptime(dstr, _date_fmt_str2)


def convert_datetime_fields(list_of_dicts, fields_to_convert):
    """
    Takes in a list of dictionaries and for each dictionary it converts all fields in fields_to_convert to
    datetime objects from Power BI Datetime Strings. This is typically used when retrieving a list of records
    from the Power BI Service and you want all date fields to be converted from the date string to python datetimes.

    If the value of the field is None or an empty string, it does nothing for that field.

    :param list_of_dicts: A list of dictionaries
    :param fields_to_convert: A list of fields to be converted to datetimes from Power BI Datetime Strings
    :return: list of dictionaries with all fields specified in 'fields_to_convert' into python datetime objects
    """
    new_list = []
    for rec in list_of_dicts:
        # Create a copy so we don't overwrite the original dictionary
        new_rec = rec.copy()
        new_list.append(new_rec)
        for field in fields_to_convert:
            if field in new_rec.keys() and new_rec[field] not in [None, '']:
                new_rec[field] = date_from_powerbi_str(new_rec[field])

    return new_list
